-- This module provides the type of natural numbers and some of its
-- properties.

open import Lib.Equality
open import Lib.Decidable
open import Lib.Proposition

module Lib.Nat where

-- ----------------------------------------------------------------------
-- Definition

-- The type of natural numbers.
data Nat : Set where
  zero : Nat
  succ : Nat -> Nat

infixl 6 _+_ _-_
infixl 7 _*_

-- Addition.
_+_ : Nat -> Nat -> Nat
zero + m = m
succ n + m = succ (n + m)

-- Truncated subtraction.
_-_ : Nat -> Nat -> Nat
n - zero = n
zero - succ m = zero
succ n - succ m = n - m

-- Multiplication.
_*_ : Nat -> Nat -> Nat
zero * m = zero
succ n * m = m + n * m

-- The following directive enables constants such as 3 to be used
-- instead of succ (succ (succ zero)).
{-# BUILTIN NATURAL Nat #-}

-- ----------------------------------------------------------------------
-- Operations

infix 5 _<=_

-- The less-than-or-equal relation.
data _<=_ : Nat -> Nat -> Set where
  leq-zero : ∀ {m} -> zero <= m
  leq-succ : ∀ {n m} -> n <= m -> succ n <= succ m

infix 5 _<'_

-- The less-than-or-equal relation.
data _<'_ : Nat -> Nat -> Set where
  less-zero : ∀ {m} -> zero <' succ m
  less-succ : ∀ {n m} -> n <' m -> succ n <' succ m

-- ----------------------------------------------------------------------
-- Properties

-- Successor is an injective function.
lemma-succ-cancellation : ∀ {n m} -> succ n == succ m -> n == m
lemma-succ-cancellation refl = refl

-- Successor is order-reflecting.
lemma-less-cancellation : ∀ {n m} -> succ n <' succ m -> n <' m
lemma-less-cancellation (less-succ x) = x

-- Successor is order-reflecting.
lemma-leq-cancellation : ∀ {n m} -> succ n <= succ m -> n <= m
lemma-leq-cancellation (leq-succ x) = x

lemma-leq-succ : ∀ {n} -> n <= succ n
lemma-leq-succ {zero} = leq-zero
lemma-leq-succ {succ n} = leq-succ lemma-leq-succ

lemma-leq-trans : ∀ {n m l} -> n <= m -> m <= l -> n <= l 
lemma-leq-trans leq-zero leq-zero = leq-zero
lemma-leq-trans leq-zero (leq-succ le2) = leq-zero
lemma-leq-trans (leq-succ le1) (leq-succ le2) = leq-succ (lemma-leq-trans le1 le2)

lemma-leq : ∀ {n m} -> succ n <= m -> succ m <= n -> False
lemma-leq {.(succ m)} {.(succ _)} (leq-succ leq1) (leq-succ {m = m} leq2) = lemma-leq leq1 leq2

lemma-n-j : ∀ {n j} -> n - j <= n
lemma-n-j {zero} {zero} = leq-zero
lemma-n-j {zero} {succ j} = leq-zero
lemma-n-j {succ n} {zero} = leq-succ (lemma-n-j {n} {zero})
lemma-n-j {succ n} {succ j} with lemma-n-j {n} {j}
... | h = lemma-leq-trans h lemma-leq-succ
 


-- ----------------------------------------------------------------------
-- Decidable properties.

infix 8 _==?_ _<=?_ _<'?_

-- Equality is decidable.
_==?_ : (n m : Nat) -> Decidable (n == m)
zero ==? zero = yes refl
zero ==? succ m = no (λ ())
succ n ==? zero = no (λ ())
succ n ==? succ m with n ==? m
succ n ==? succ m | yes w = yes (context succ w)
succ n ==? succ m | no w = no (λ k → w (lemma-succ-cancellation k))

-- Leq is decidable.
_<=?_ : (n m : Nat) -> Decidable (n <= m)
zero <=? m = yes leq-zero
succ n <=? zero = no (λ ())
succ n <=? succ m with n <=? m
succ n <=? succ m | yes w = yes (leq-succ w)
succ n <=? succ m | no w = no (λ k → w (lemma-leq-cancellation k))

-- Less is decidable.
_<'?_ : (n m : Nat) -> Decidable (n <' m)
zero <'? zero = no (λ ())
zero <'? succ m = yes less-zero
succ n <'? zero = no (λ ())
succ n <'? succ m with n <'? m
... | yes x = yes (less-succ x)
... | no x = no (λ k → x (lemma-less-cancellation k))
