-- This module provides the definition of words in some generators (in
-- the language of monoids), as well as the axioms for reasoning about
-- equality of such words.

open import Lib.Nat
open import Lib.Proposition
open import Lib.Equality as Eq using (_==_)
open import Lib.Inspect
open import Lib.List

module Lib.Word where

-- ----------------------------------------------------------------------
-- * Precedence of various operators

infix 8 _^_
infixr 7 _•_
infix 6 _===_
infix 5 _⊢_
infix 5 _∈_

-- ----------------------------------------------------------------------
-- * Words

-- Words in the language of monoids, over a set of generators X. A
-- word is either a generator, the empty word, or a product of two
-- words.
data Word (X : Set) : Set where
  gen : X -> Word X
  ε : Word X
  _•_ : Word X -> Word X -> Word X

-- For convenience, we define powers of words.
_^_ : ∀ {X} -> Word X -> Nat -> Word X
w ^ zero = ε
w ^ (succ zero) = w
w ^ succ (succ n) = w • (w ^ (succ n))

-- ----------------------------------------------------------------------
-- * Relations

-- A relation is just a pair of words, written v === w.
data Relation (X : Set) : Set where
  _===_ : Word X -> Word X -> Relation X

-- A context is a set of relations.
Context : Set -> Set₁
Context X = Relation X -> Set

-- The empty context.
∅ : ∀ {X} -> Context X
∅ E = False

-- Notation: we write E ∈ Γ if the relation E is in the context Γ.
_∈_ : ∀ {X} -> Relation X -> Context X -> Set
E ∈ Γ = Γ E

-- Syntactic equational reasoning. Judgements are of the form Γ ⊢ E,
-- where Γ is a context and E is a relation. A judgement is valid if
-- the relations in Γ imply the relation E. Valid judgements are
-- derived by the following rules.
data _⊢_ {X : Set} (Γ : Context X) : Relation X -> Set where
  -- Any relation in Γ is true by assumption.
  axiom : ∀ {E} -> E ∈ Γ -> Γ ⊢ E

  -- The relation '===' is a congruence.
  refl : ∀ {w} -> Γ ⊢ w === w
  symm : ∀ {w v} -> Γ ⊢ w === v -> Γ ⊢ v === w
  trans : ∀ {w v u} -> Γ ⊢ w === v -> Γ ⊢ v === u -> Γ ⊢ w === u
  cong : ∀ {w w' v v'} -> Γ ⊢ w === w' -> Γ ⊢ v === v' -> Γ ⊢ w • v === w' • v'

  -- The monoid axioms: associativity and the left and right unit laws.
  assoc : ∀ {w v u} -> Γ ⊢ (w • v) • u === w • (v • u)
  left-unit : ∀ {w} -> Γ ⊢ ε • w === w
  right-unit : ∀ {w} -> Γ ⊢ w • ε === w

-- ----------------------------------------------------------------------
-- * Auxiliary operations

-- Extend a function from generators to words.
_* : {X Y : Set} -> (X -> Word Y) -> (Word X -> Word Y)
_* g (gen x) = g x
_* g ε = ε
_* g (w • u) = (g *) w • (g *) u

-- ----------------------------------------------------------------------
-- * Definition of soundness and completeness.

_is-sound-wrt_and_ : ∀ {X Y} -> Context X -> Context Y -> (f : X -> Word Y) -> Set
_is-sound-wrt_and_ {X} {Y} Γ Δ f = ∀ {w v : Word X} -> Γ ⊢ w === v -> Δ ⊢ (f *) w === (f *) v

_is-complete-wrt_and_ : ∀ {X Y} -> Context X -> Context Y -> (f : X -> Word Y) -> Set
_is-complete-wrt_and_ {X} {Y} Γ Δ f = ∀ {w v : Word X} -> Δ ⊢ (f *) w === (f *) v -> Γ ⊢ w === v


-- Map for word.
wmap : ∀ {a b} -> (f : a -> b) -> Word a -> Word b
wmap f (gen x) = gen (f x)
wmap f ε = ε
wmap f (wa • wa₁) = wmap f wa • wmap f wa₁

infix 10 _⌟
infix 9 ⌞_

_⌟ : ∀ {a b} -> Word b -> Word (a ∪ b)
x ⌟ = wmap _⌝ x 

⌞_ : ∀ {a b} -> Word a -> Word (a ∪ b)
⌞ x = wmap ⌜_ x 

infix 10 _◹
infix 9 ◸_

◸_ : ∀ {A B} -> (e : Relation A) -> Relation (A ∪ B)
◸ (x === x₁) = ⌞ x === ⌞ x₁

_◹ : ∀ {A B} -> (e : Relation B) -> Relation (A ∪ B)
_◹ (x === x₁) = x ⌟ === x₁ ⌟

infix 9 ⌜_
infix 10 _⌝
infix 10 ◃_▹
infix 10 ◂_▸

infixr 10 _⊹_⊹_

data _⊹_⊹_ {A B} (Γ₁ : Context A) (Γ₂ : Context B) (Γ : Context (A ∪ B)) : Context (A ∪ B) where
  ⌜_ :  ∀ {u v} -> u === v ∈ Γ₁ -> ◸ (u === v) ∈ Γ₁ ⊹ Γ₂ ⊹ Γ
  _⌝ :  ∀ {u v} -> u === v ∈ Γ₂ -> (u === v) ◹ ∈ Γ₁ ⊹ Γ₂ ⊹ Γ
  ◃_▹ :  ∀ {u v} -> u === v ∈ Γ -> u === v ∈ Γ₁ ⊹ Γ₂ ⊹ Γ



infix 10 _⌋
infix 9 ⌊_

⌊_ : ∀  {A B} {Γ₁ : Context A } {Γ₂ : Context B} {Γ : Context (A ∪ B)} {u v} -> Γ₁ ⊢ u === v -> Γ₁ ⊹ Γ₂ ⊹ Γ ⊢ ◸ (u === v)
⌊ (axiom x) = axiom (⌜ x)
⌊ refl = refl
⌊ (symm d) = symm (⌊ d)
⌊ (trans d d₁) = trans (⌊ d) (⌊ d₁)
⌊ (cong d d₁) = cong (⌊ d) (⌊ d₁)
⌊ assoc = assoc
⌊ left-unit = left-unit
⌊ right-unit = right-unit

_⌋ : ∀  {A B} {Γ₁ : Context A } {Γ₂ : Context B} {Γ : Context (A ∪ B)} {u v} -> Γ₂ ⊢ u === v -> Γ₁ ⊹ Γ₂ ⊹ Γ ⊢ (u === v) ◹
_⌋ (axiom x) = axiom (x ⌝)
_⌋ refl = refl
_⌋ (symm d) = symm (_⌋ d)
_⌋ (trans d d₁) = trans (_⌋ d) (_⌋ d₁)
_⌋ (cong d d₁) = cong (_⌋ d) (_⌋ d₁)
_⌋ assoc = assoc
_⌋ left-unit = left-unit
_⌋ right-unit = right-unit

◂_▸ : ∀  {A B} {Γ₁ : Context A } {Γ₂ : Context B} {Γ : Context (A ∪ B)} {u v} -> Γ ⊢ u === v -> Γ₁ ⊹ Γ₂ ⊹ Γ ⊢ u === v
◂_▸ (axiom x) = axiom ◃ x ▹
◂_▸ refl = refl
◂_▸ (symm d) = symm (◂_▸ d)
◂_▸ (trans d d₁) = trans (◂_▸ d) (◂_▸ d₁)
◂_▸ (cong d d₁) = cong (◂_▸ d) (◂_▸ d₁)
◂_▸ assoc = assoc
◂_▸ left-unit = left-unit
◂_▸ right-unit = right-unit

